/*
  OMPi OpenMP Compiler
  == Copyright since 2001 the OMPi Team
  == Department of Computer Science, University of Ioannina

  This file is part of OMPi.

  OMPi is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  OMPi is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with OMPi; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* ort_tasks.c -- OMPi RunTime library, tasking */

/*
 * 2010/11/20:
 *   added calls for getting & setting task icvs..
 * Version 1.0.1j:
 *   first time around, out of ort.c code.
 */
 
#include "ort_prive.h"
#include <stdlib.h>
#include <stdio.h>

#define FAILURE 0
#define SUCCESS 1


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                   *
 * GLOBAL VARIABLES / DEFINITIONS / MACROS                           *
 *                                                                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
                                     

/* Per-thread task throttling */
#ifdef USE_TLS
  TLS_KEYWORD int throttle;  
  #define __start_throttling() (throttle = 1)
  #define __stop_throttling()  (throttle = 0)
  #define __check_throttling() (throttle + 0)
#else
  ee_key_t   throttle_key;    /* For thread-specific task throttling */
  #define __start_throttling() ee_setspecific(throttle_key,(void *) 1)
  #define __stop_throttling()  ee_setspecific(throttle_key,(void *) 0)
  #define __check_throttling() (0 != (int) ee_getspecific(throttle_key))
#endif

ort_task_node_t *ort_task_alloc(ort_eecb_t *thr, void *arg);
void             ort_task_free(ort_eecb_t *thr, ort_task_node_t *node);


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                   *
 * TASKS                                                             *
 *                                                                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
                                     

void ort_init_tasking()
{
#ifdef USE_TLS
#else
  ee_key_create(&throttle_key, 0);
#endif
  __stop_throttling();
}


/* Allocate memory for my group task queues */
void ort_task_queues_init(ort_eecb_t *me, int nthr)
{
  int i;
  
  /* Thread becomes parent for first time */
  if (me->tasking.queue_table == NULL)
  {
    me->tasking.queue_table = (ort_task_queue_t*)
                              ort_calloc((nthr + 1)*sizeof(ort_task_queue_t));

    for (i=0; i<nthr + 1; i++)   /* Initialize task queues */
    {
      me->tasking.queue_table[i].top         = 0;
      me->tasking.queue_table[i].bottom      = 0;
      me->tasking.queue_table[i].implicit_task_children = NULL;
      #if !defined(HAVE_ATOMIC_FAA) || !defined(HAVE_ATOMIC_CAS)
        ee_init_lock((ee_lock_t *) &(me->tasking.queue_table[i].lock),
                     ORT_LOCK_NORMAL);
      #endif
      
      
    }
    me->tasking.max_children = nthr + 1;
  }
  else 
  {
    if (me->tasking.max_children < nthr + 1)  /* realloc needed */
    {
      /* Reallocate queue_tabel */
      me->tasking.queue_table = (ort_task_queue_t*)
         realloc(me->tasking.queue_table, (nthr + 1)*sizeof(ort_task_queue_t));

      #if !defined(HAVE_ATOMIC_FAA) || !defined(HAVE_ATOMIC_CAS)
      for (i=me->tasking.max_children; i<nthr + 1; i++)
        ee_init_lock((ee_lock_t *) &(me->tasking.queue_table[i].lock),
                     ORT_LOCK_NORMAL);
      #endif
      
      me->tasking.max_children = nthr + 1;
    }
    for (i=0; i<nthr + 1; i++)
    {
      me->tasking.queue_table[i].top      = 0;
      me->tasking.queue_table[i].bottom   = 0;
      me->tasking.queue_table[i].implicit_task_children = NULL;
    }
  }
}

/* ort_task_alloc
 * This function tries to use the recycle bin of process 'thread_id'
 * in order to reuse a previusly used 'ort_task_node_t' structure. if recycle
 * bin is empty then a new 'ort_task_node_t' is returned using malloc.
 */
ort_task_node_t* ort_task_alloc(ort_eecb_t *thr, void *arg)
{
  int i;
  ort_task_node_t *new_node;

  if(arg != NULL)
  {
    new_node = (ort_task_node_t*) (*((void**)(((char*)arg) - 2*sizeof(void*))));
    if(new_node != NULL)
    {
      new_node->occupied = 1;
      return new_node;
    }
  }
  
  if(thr->tasking.task_node_pool != NULL)
  {
    for(i=0; i<TASKPOOLSIZE(thr->num_siblings); i++)
      if(thr->tasking.task_node_pool[i].occupied == 0)
      {
	thr->tasking.task_node_pool[i].occupied = 1;
	return &(thr->tasking.task_node_pool[i]);
      };
  }

  new_node = thr->tasking.task_node_recycler;
  if(new_node == NULL)
    new_node = (ort_task_node_t*)ort_alloc(sizeof(ort_task_node_t));
  else
    thr->tasking.task_node_recycler = new_node->next;

  new_node->occupied = 2;
  return new_node;
}


/* ort_task_free
 * This function puts a 'ort_task_node_t' structure in process's 'thread_id'
 * recycle bin in order to reuse it in the future.
 */
void ort_task_free(ort_eecb_t *thr, ort_task_node_t* node)
{
  if(node->occupied == 1)
    node->occupied = 0;
  else
  {
    node->next = thr->tasking.task_node_recycler;
    thr->tasking.task_node_recycler = node;
  }
}


/* ort_task_worker_enqueue
 * This function is used by a process (worker) in order to enqueue a new
 * task_node to its task queue.
 */
static int ort_task_worker_enqueue(int worker_id, void* (*func)(void *),
                                   void *arg)
{

    ort_task_node_t *new_node;
    ort_eecb_t *me = __MYCB;
    ort_eecb_t *my_parent = me->sdn;
    int old_bottom = atomic_read
                       (&(my_parent->tasking.queue_table[worker_id].bottom));
    int old_top = atomic_read
                       (&(my_parent->tasking.queue_table[worker_id].top));
    int size = old_bottom - old_top;

    if (size >= TASKQUEUESIZE) /* Task queue full */
      return FAILURE;
    
    new_node = ort_task_alloc(me, arg);

    #if !defined(HAVE_ATOMIC_FAA)
      ee_init_lock((ee_lock_t *) &(new_node->lock), ORT_LOCK_NORMAL); 
    #endif
    new_node->func         = func;
    new_node->funcarg      = arg;
    new_node->num_children = 0;
    new_node->next         = NULL;
    new_node->parent       = __CURRTASK(me);
    new_node->icvs         = new_node->parent->icvs;

    if(new_node->parent != NULL)   /* Add a new child to parent task */
    {
      #if defined(HAVE_ATOMIC_FAA)
        _faa(&((new_node->parent)->num_children),1);
      #else
        ee_set_lock(&((new_node->parent)->lock));
        (new_node->parent)->num_children++;
        ee_unset_lock(&((new_node->parent)->lock));
      #endif
    }
    my_parent->tasking.queue_table[worker_id].tasks[old_bottom%TASKQUEUESIZE] =
                                                                       new_node;

    #if defined(HAVE_ATOMIC_FAA)
      _faa(&(my_parent->tasking.queue_table[worker_id].bottom),1);
    #else
      ee_set_lock(&(my_parent->tasking.queue_table[worker_id].lock));
      my_parent->tasking.queue_table[worker_id].bottom++;
      ee_unset_lock(&(my_parent->tasking.queue_table[worker_id].lock));
    #endif

    #ifdef ORT_DEBUG
      me->tasking.tasks_enqueued++;
    #endif

    return SUCCESS;
}


/* ort_task_worker_dequeue
 * This function is used by a process (worker) in order to dequeue a
 * new task_node from its task queue.
 */
static ort_task_node_t*  ort_task_worker_dequeue(int worker_id)
{
    ort_task_node_t* extracting_node;
    ort_eecb_t *my_parent = __MYCB->sdn;
    int old_top;
    int new_top;
    int size;

    size = atomic_read(&(my_parent->tasking.queue_table[worker_id].bottom))
	 - atomic_read(&(my_parent->tasking.queue_table[worker_id].top)) -1;
    
    if (size < 0) /* Queue is empty */
      return NULL;
    
    #if defined(HAVE_ATOMIC_FAA)
      _faa(&(my_parent->tasking.queue_table[worker_id].bottom),-1);
    #else
      ee_set_lock(&(my_parent->tasking.queue_table[worker_id].lock));
      my_parent->tasking.queue_table[worker_id].bottom--;
      ee_unset_lock(&(my_parent->tasking.queue_table[worker_id].lock));
    #endif

    old_top = atomic_read(&(my_parent->tasking.queue_table[worker_id].top));
    new_top = old_top + 1;
    size = atomic_read(&(my_parent->tasking.queue_table[worker_id].bottom))
               - old_top;

    if (size < 0) /* Queue is empty */
    {
        my_parent->tasking.queue_table[worker_id].bottom = old_top;
        return NULL;
    }

    extracting_node = my_parent->tasking.queue_table[worker_id].tasks\
                             [atomic_read(&((my_parent->tasking.queue_table\
                             [worker_id]).bottom))%TASKQUEUESIZE];
    if(size > 0)
        return extracting_node;

    /* If there is only one task left in queue... */
    #if defined(HAVE_ATOMIC_CAS)
      /* If a thief stole the last task... */
      if(!_cas(&((my_parent->tasking.queue_table[worker_id]).top)\
                                                , old_top, new_top))
        extracting_node = NULL;/* then return NULL, else return the last task */
    #else
      ee_set_lock(&((my_parent->tasking.queue_table[worker_id]).lock));

      if((my_parent->tasking.queue_table[worker_id]).top == old_top)
        (my_parent->tasking.queue_table[worker_id]).top = new_top;
      else
        extracting_node = NULL;

      ee_unset_lock(&((my_parent->tasking.queue_table[worker_id]).lock));
    #endif

    /* old_top + 1 = new_top */
    my_parent->tasking.queue_table[worker_id].bottom = old_top + 1;

    return extracting_node;
}

/* ort_task_thief_steal
 * This function is used by a process (thief) in order to dequeue a new
 * task_node from a victim process's task queue.
 */

static ort_task_node_t* ort_task_thief_steal(ort_eecb_t *me, int victim_id)
{
    ort_task_node_t *extracting_node;
    ort_eecb_t *my_parent = me->sdn;
    int old_top = atomic_read\
                          (&(my_parent->tasking.queue_table[victim_id].top));
    int new_top = old_top + 1;
    int old_bottom = atomic_read\
                       (&(my_parent->tasking.queue_table[victim_id].bottom));
    int size = old_bottom - old_top;

    if(size <= 0) /* Victim's queue is empty */
        return NULL;
    /* Steal a task from vitim's top! */
    extracting_node = my_parent->tasking.queue_table\
                                            [victim_id].tasks[old_top%TASKQUEUESIZE];

    #if defined(HAVE_ATOMIC_CAS)
      /* if thief managed to steal the task... */
      if(_cas(&((my_parent->tasking.queue_table\
                                           [victim_id]).top), old_top, new_top))
        return extracting_node;
    #else
      ee_set_lock(&((my_parent->tasking.queue_table[victim_id]).lock));
      if((my_parent->tasking.queue_table[victim_id]).top == old_top)
      {
        (my_parent->tasking.queue_table[victim_id]).top = new_top;
        ee_unset_lock(&((my_parent->tasking.queue_table[victim_id]).lock));
        return extracting_node;
      }
      ee_unset_lock(&((my_parent->tasking.queue_table[victim_id]).lock));
    #endif

    return NULL;
}


/* ort_task_execute
 * This function is part of openmp interface. When a thread calls it,
 * it dequeues a task from its queue, or steals a task from a
 * victim and executes it.
 */
static int ort_task_execute(ort_eecb_t *me, int startfrom)
{
    int thread_id = me->thread_num;
    ort_eecb_t *my_parent = me->sdn;
    ort_task_node_t* task_to_execute = ort_task_worker_dequeue(thread_id);
    ort_task_node_t* prev_task_to_execute;
    int victim;

    /* Processes task queue is empty! Have to steal something... */
    if(task_to_execute == NULL)
    {
        if (startfrom > 0)              /* Start stealing from there */
	  thread_id = startfrom - 1;
	
        for(victim=(thread_id+1);victim<(my_parent->num_children+thread_id);victim++)
        {
            task_to_execute = ort_task_thief_steal\
                                             (me, victim%(my_parent->num_children));
             /* If victim's queue is also empty, try next victim */
            if(task_to_execute == NULL)
                    continue;

            /* Finally i stole something and i am about to execute it! */
            /* First update thread executing status... */
            prev_task_to_execute = __CURRTASK(me);
	    __CURRTASK(me) = task_to_execute;

            (task_to_execute->func)(task_to_execute->funcarg);

            /* Finished my job. Update thread executing status */
            __CURRTASK(me) = prev_task_to_execute;
	    
            if(task_to_execute->parent != NULL)  /* If task has a parent then */
            {
               #if defined(HAVE_ATOMIC_FAA)
                 _faa(&((task_to_execute->parent)->num_children), -1);
               #else
                 ee_set_lock(&((task_to_execute->parent)->lock));
                 (task_to_execute->parent)->num_children--;
                 ee_unset_lock(&((task_to_execute->parent)->lock));
              #endif
            }
	    #ifdef ORT_DEBUG
	      me->tasking.tasks_executed_by_thief++;
	    #endif
            /* Free memory */
            ort_task_free(me, task_to_execute); /* Recycle task node */

            return SUCCESS; /* I succesfully stole and executed a task! */
        }
        return FAILURE; /* There was no task left to execute */
    }
    else
    {
        /*Finally i am about to execute a task! */
        /* First update thread executing status... */
        prev_task_to_execute = __CURRTASK(me);
        __CURRTASK(me) = task_to_execute;
	
        (task_to_execute->func)(task_to_execute->funcarg);

        /* Finished my job. Update thread executing status */
        __CURRTASK(me) = prev_task_to_execute;

        /* If task has a parent then */
        if(task_to_execute->parent != NULL)
        {
          #if defined(HAVE_ATOMIC_FAA)
            _faa(&((task_to_execute->parent)->num_children), -1);
          #else
            ee_set_lock(&((task_to_execute->parent)->lock));
            (task_to_execute->parent)->num_children--;
            ee_unset_lock(&((task_to_execute->parent)->lock));
          #endif
        }

	#ifdef ORT_DEBUG
	  me->tasking.tasks_executed_by_worker++;
	#endif
        /* Free memory */
        ort_task_free(me, task_to_execute); /* Recycle task node */
        return SUCCESS; /* I succesfully executed a task! */
    }
}


/* Only execute tasks from my queue (i.e. do not attempt to steal) 
 */
void ort_execute_my_tasks(ort_eecb_t *me)
{
  ort_task_node_t *task, *currtask = __CURRTASK(me);
  
  for (;;)
  {
    task = ort_task_worker_dequeue(me->thread_num);
    if (task == NULL)     /* drained */
      return;
    
    __CURRTASK(me) = task;
    (task->func)(task->funcarg);
    __CURRTASK(me) = currtask;

    if (task->parent != NULL)
    {
      #if defined(HAVE_ATOMIC_FAA)
	_faa(&(task->parent->num_children), -1);
      #else
	ee_set_lock(&(task->parent->lock));
	task->parent->num_children--;
	ee_unset_lock(&(task->parent->lock));
      #endif
    }

    #ifdef ORT_DEBUG
      me->tasking.tasks_executed_by_worker++;
    #endif
    /* Free memory */
    ort_task_free(me, task); /* Recycle task node */
  }
}

/* This function creates a new task node and sets this task as the
 * calling thread's current task. Called when a task is about to be 
 * executed immediately.
 */
void *ort_task_immediate_start(void)
{
  ort_eecb_t      *me = __MYCB;
  ort_task_node_t *new_node;
  
  #ifdef ORT_DEBUG
    me->tasking.throttled++;
  #endif
  
  /* Create a task node in order to save task data */
  new_node = ort_task_alloc(me, NULL);
  
  #if !defined(HAVE_ATOMIC_FAA)
    ee_init_lock((ee_lock_t *) &(new_node->lock), ORT_LOCK_NORMAL); 
  #endif
  new_node->func         = NULL;
  new_node->funcarg      = NULL;
  new_node->num_children = 0;
  new_node->next         = NULL;
  new_node->parent       = __CURRTASK(me);
  new_node->icvs         = __CURRTASK(me)->icvs;
  
  __CURRTASK(me) = new_node;
  
  return (void *) new_node;
}


void ort_task_immediate_end(void *new_node)
{
  ort_eecb_t      *me = __MYCB;
  ort_task_node_t *task_node = (ort_task_node_t *) new_node;
  
  __CURRTASK(me) = task_node->parent;    /* Restore task node */
  ort_task_free(me, task_node);          /* Recycle task node */
}


/* Create a new task
 */
void ort_new_task(int untied, void *(*func)(void *arg), void *arg)
{
#if defined(AVOID_OMPI_DEFAULT_TASKS)

  ee_new_task(untied, func, arg);
  
#else
  ort_eecb_t      *me = __MYCB;
  ort_task_node_t *tnode;

  if (arg != NULL)
    if(((ort_task_env_subpool_t*)(((char*)arg)-sizeof(void*)))->occupied == 2 ||
       ((ort_task_env_subpool_t*)(((char*)arg)-sizeof(void*)))->occupied == 3)
    {
      #ifdef ORT_DEBUG
	me->tasking.throttled_pool++;
      #endif
      
      EXECUTE_NOW:
        __start_throttling();
        tnode = ort_task_immediate_start();
        (*func)(arg);
        ort_task_immediate_end(tnode);
        __stop_throttling();
        return;
    }

  if (me->sdn == NULL || me->num_siblings == 1)
  {
    #ifdef ORT_DEBUG
      me->tasking.throttled_serial++;
    #endif
    goto EXECUTE_NOW;
  }

  if (ort_task_worker_enqueue(me->thread_num, func, arg) == FAILURE)
  {
    #ifdef ORT_DEBUG
      me->tasking.throttled_queue++;
    #endif 
    goto EXECUTE_NOW;
  }
  else
  {
    me->parent->tasking.never_task = 1;
    FENCE;
    __stop_throttling();
  }
#endif
}


/* Create a new task & execute it immediately
 */
void ort_new_task_exec(int untied, void *(*func)(void *arg), void *arg)
{
#if defined(AVOID_OMPI_DEFAULT_TASKS)
  ee_new_task_exec(untied, func, arg);
#else
  ort_task_node_t *new_node;
  
  #ifdef ORT_DEBUG
    __MYCB->tasking.throttled_if++;
  #endif

  new_node = ort_task_immediate_start();
  (*func)(arg);
  ort_task_immediate_end(new_node);
#endif
}


inline
int check_for_tasks(ort_eecb_t *me)
{
  int              teamsize = me->num_siblings, victim, retry = 0;
  ort_task_queue_t *q = me->parent->tasking.queue_table;
  
  /* Search for # of unfinished tasks in my mates queues */
  for (victim = me->thread_num+1; victim < teamsize + me->thread_num; victim++)
  {
    if ((q[victim % teamsize]).implicit_task_children == NULL)
      retry = -1;
    else
      if (*((q[victim % teamsize]).implicit_task_children) > 0)
        return ( victim % teamsize );
  }
  return retry;
}


/* How = 0 (wait for my children), 1 (wait for all team tasks),
 *       2 (wait at the end of parallel)
 */
void ort_taskwait(int how)
{
#if defined(AVOID_OMPI_DEFAULT_TASKS)
  ee_taskwait(how);
#else
  ort_eecb_t *me = __MYCB;
  int        victim = -1;

  if (me->parent == NULL || me->parent->tasking.never_task == 0
                         || me->num_siblings == 1) 
    return;

  if (how > 0)   /* help with all the tasks in current team */
  {
    do{
      while(ort_task_execute(me, victim));
    }while ((victim = check_for_tasks(me)) != 0);
  }
  else           /* execute till all my child tasks finish */
    while (__CURRTASK(me)->num_children > 0)
      ort_task_execute(me, victim);
#endif
}


/* Task throttling.
 * For the moment, this is a per-thread flag that should be adjusted
 * adaptively. 
 * A simple policy would be to __start_throttling() when the number of 
 * tasks in my private queue exceeds c*N where c is a constant and N 
 * is the number of processors. If later I discover that the number
 * fell below this threshold, I __stop_throttling().
 * 
 */
int ort_task_throttling()
{
  return __check_throttling();
}


void *ort_taskenv_alloc(int size)
{
#if defined(AVOID_OMPI_DEFAULT_TASKS)
  return ( malloc(size) );
#else
  ort_eecb_t      *me = __MYCB;
  ort_task_env_pool_t *task_env_pool;
  ort_task_env_subpool_t *task_env_subpool;
  ort_task_node_t* task_node_pool;
  int i;
  
  task_env_pool = me->tasking.task_env_pool;
  task_node_pool = me->tasking.task_node_pool;

  while(task_env_pool != NULL)
  {
    if(task_env_pool->bin_size == size) /* I found matching pool*/
    {
      task_env_subpool = task_env_pool->sub_pool;

      if(task_node_pool != NULL)
      {
	for(i=0; i<(TASKPOOLSIZE(me->num_siblings)-1); i++)
	  if (task_node_pool[i].occupied == 0)
	  {
	    if(task_env_pool->sub_pool[i].data == NULL)
	    {
	      task_env_pool->sub_pool[i].data = ort_alloc(size+2*sizeof(void*));
	      *((void **)(task_env_pool->sub_pool[i].data)) = 
	         &(task_node_pool[i]);
	      *((void **)(((char*) task_env_pool->sub_pool[i].data)+sizeof(void*))) = 
	         &(task_env_pool->sub_pool[i]);
	    }
	    task_env_subpool[i].occupied = 1;
	    return (void*) (((char*)task_env_subpool[i].data)+2*sizeof(void*));
	  }
	/* Pool is about to be full allocate last element in pool */
	if(task_node_pool[i].occupied == 0)
	{
	  if(task_env_pool->sub_pool[i].data == NULL)
	  {
	    task_env_pool->sub_pool[i].data = ort_alloc(size+2*sizeof(void*));
	    *((void **)(task_env_pool->sub_pool[i].data)) = 
	       &(task_node_pool[i]);
	    *((void **)(((char*)task_env_pool->sub_pool[i].data)+sizeof(void*))) = 
	       &(task_env_pool->sub_pool[i]);
	  }
	  task_env_subpool[i].occupied = 2;
	  return (void*) (((char*) task_env_subpool[i].data) + 2*sizeof(void*));
	}
	else
	  goto ALLOCATE_FROM_RECYCLE;
      }
      else
      {
ALLOCATE_FROM_RECYCLE:
	task_env_subpool = me->tasking.task_env_pool->recycler;
	if(task_env_subpool == NULL)
	{
	  task_env_subpool = (ort_task_env_subpool_t*)\
	                      ort_calloc(sizeof(ort_task_env_subpool_t)); 
	  task_env_subpool->data = ort_alloc(size+2*sizeof(void*));
	  task_env_subpool->next = NULL;
	  *((void **)(task_env_subpool->data)) = NULL;
	  *((void **)(((char*)task_env_subpool->data)+sizeof(void*))) = 
	    task_env_subpool;
	  task_env_subpool->occupied = 3;
	}
	else
	  me->tasking.task_env_pool->recycler = task_env_subpool->next;
	
	return (void*) (((char*) task_env_subpool->data) + 2*sizeof(void*));
      }
    }
    task_env_pool = task_env_pool->next;
  }

  /* I did not find a matching rbin. allocate some space */
  task_env_pool = (ort_task_env_pool_t *)ort_calloc(sizeof(ort_task_env_pool_t));
  task_env_pool->bin_size = size;
  task_env_pool->recycler = NULL;
  task_env_pool->sub_pool = (ort_task_env_subpool_t*)\
			  ort_calloc(TASKPOOLSIZE(me->num_siblings)*sizeof(ort_task_env_subpool_t)); 

  task_env_pool->next = me->tasking.task_env_pool;
  me->tasking.task_env_pool = task_env_pool;

  task_env_subpool = task_env_pool->sub_pool;
  if(task_node_pool != NULL)
  {
    for(i=0; i<(TASKPOOLSIZE(me->num_siblings)-1); i++)
      if (task_node_pool[i].occupied == 0)
      {
	if(task_env_pool->sub_pool[i].data == NULL)
	{
	  task_env_pool->sub_pool[i].data = ort_alloc(size+2*sizeof(void*));
	  *((void **)(task_env_pool->sub_pool[i].data)) = &(task_node_pool[i]);
	  *((void **)(((char*)task_env_pool->sub_pool[i].data)+sizeof(void*))) = 
	    &(task_env_pool->sub_pool[i]);
	}
	task_env_subpool[i].occupied = 1;
	return (void*) (((char*)task_env_subpool[i].data) + 2*sizeof(void*));
      }
    /* Pool is about to be full allocate last element in pool */
    if(task_node_pool[i].occupied == 0)
    {
      if(task_env_pool->sub_pool[i].data == NULL)
      {
	task_env_pool->sub_pool[i].data = ort_alloc(size+2*sizeof(void*));
	*((void **)(task_env_pool->sub_pool[i].data)) = &(task_node_pool[i]);
	*((void **)(((char*)task_env_pool->sub_pool[i].data)+sizeof(void*))) = 
	  &(task_env_pool->sub_pool[i]);
      }
      task_env_subpool[i].occupied = 2;
      return (void*) (((char*)task_env_subpool[i].data) + 2*sizeof(void*));
    }
    else goto ALLOCATE_FROM_NEW_RECYCLE;
  }
  else
  {
ALLOCATE_FROM_NEW_RECYCLE:
    task_env_subpool = (ort_task_env_subpool_t*)ort_calloc(sizeof(ort_task_env_subpool_t)); 
    task_env_subpool->data = (void *)ort_alloc(size+2*sizeof(void*));
    task_env_subpool->next = NULL;
    *((void **)(task_env_subpool->data)) = NULL;
    *((void **)(((char*)task_env_subpool->data)+sizeof(void*))) = task_env_subpool;

    task_env_subpool->occupied = 3;
    return (void*) (((char*)task_env_subpool->data) + 2*sizeof(void*));
  }
  
  printf("ERROR IN TASK ENV ALLOCATE\n");fflush(stdout);
  return NULL;
#endif
}

void  ort_taskenv_free(void *ptr, int size)
{
#if defined(AVOID_OMPI_DEFAULT_TASKS)
  free(ptr);

  return;
#else
  ort_eecb_t      *me = __MYCB;
  ort_task_env_subpool_t *recycling_node = (ort_task_env_subpool_t*)
                            (*((void **)(((char*)ptr) - sizeof(void*))));
  ort_task_env_pool_t *task_env_pool;
  
  if(recycling_node->occupied == 1 || recycling_node->occupied == 2)
    recycling_node->occupied = 0;

  else if(recycling_node->occupied == 3)
  {
    /* Search for appropriate recycle bin */
    task_env_pool = me->tasking.task_env_pool;
    while(task_env_pool != NULL)
    {
      if(task_env_pool->bin_size == size) /* I found matching bin*/
      {
	recycling_node->next = task_env_pool->recycler;
	task_env_pool->recycler = recycling_node;
	return;
      }
      task_env_pool = task_env_pool->next;
    }
    
    /* Recycle bin not found. Create a new task environment pool and a new recycle bin */
    task_env_pool = (ort_task_env_pool_t *)ort_calloc(sizeof(ort_task_env_pool_t));
    task_env_pool->bin_size = size;
    task_env_pool->recycler = recycling_node;
    task_env_pool->sub_pool = (ort_task_env_subpool_t*)\
			    ort_calloc(TASKPOOLSIZE(me->num_siblings)*sizeof(ort_task_env_subpool_t)); 

    task_env_pool->next = me->tasking.task_env_pool;
    me->tasking.task_env_pool = task_env_pool;
  }

  return;
#endif
}


/* Only called from othr.c, when in nestable locks */
void *ort_get_current_task()
{
  return (void *) __CURRTASK(__MYCB);
}


inline
void task_pools_init(ort_eecb_t *t)
{
  ort_tasking_t       *td = &(t->tasking);
  ort_task_env_pool_t *tep;
  int                 i, teamsize = t->num_siblings;
  
  if (td->task_node_pool == NULL)    /* 1st time around as father */
  {
    td->task_node_pool = (ort_task_node_t *)
         ort_calloc(TASKPOOLSIZE(teamsize)*sizeof(ort_task_node_t));
    td->max_mates = teamsize;
    return;
  }

  if (td->max_mates >= teamsize)
    return;
  
  td->task_node_pool = (ort_task_node_t *)
    ort_realloc(td->task_node_pool, 
		TASKPOOLSIZE(teamsize)*sizeof(ort_task_node_t));
  
  for (i = 0; i < TASKPOOLSIZE(teamsize); i++)
    td->task_node_pool[i].occupied = 0;
  
  for (tep = td->task_env_pool; tep != NULL; tep = tep->next)
  {
    for (i = 0; i < TASKPOOLSIZE(td->max_mates); i++)
      if (tep->sub_pool[i].data != NULL)
        free(tep->sub_pool[i].data);
    tep->sub_pool = (ort_task_env_subpool_t*)
      ort_realloc(tep->sub_pool,
                  TASKPOOLSIZE(teamsize)*sizeof(ort_task_env_subpool_t));
    for (i = 0; i< TASKPOOLSIZE(teamsize); i++)
      tep->sub_pool[i].data = NULL;
  }

  td->max_mates = teamsize;
}


void ort_start_implicit_task(ort_eecb_t *thr)
{
  ort_eecb_t      *parent_thread = thr->parent;
  ort_task_node_t *tnode, *parent_task = __CURRTASK(parent_thread);
  
  task_pools_init(thr);
   
  tnode = ort_task_alloc(thr, NULL);
  tnode->func         = parent_thread->workfunc;
  tnode->funcarg      = (void *) thr;
  tnode->num_children = 0;
  tnode->next         = NULL;
  tnode->parent       = parent_task;
  tnode->icvs         = parent_task->icvs;

  /* Save # of children in order to use it in barrier task wait */
  (parent_thread->tasking.queue_table[thr->thread_num]).implicit_task_children
     = &(tnode->num_children);

  #if defined(HAVE_ATOMIC_FAA)
    _faa(&(parent_task->num_children), 1);
  #else
    ee_set_lock(&(parent_task->lock));
    (parent_task->num_children)++;
    ee_unset_lock(&(parent_task->lock));
  #endif
  
  __CURRTASK(thr) = tnode;
}


void ort_finish_implicit_task(ort_eecb_t *thr)
{
  ort_task_node_t *tnode = __CURRTASK(thr);
  ort_eecb_t      *parent_thread = thr->parent;

  #if defined(HAVE_ATOMIC_FAA)
    _faa(&(tnode->parent->num_children), -1);
  #else
    ee_set_lock(&(tnode->parent->lock));
    tnode->parent->num_children--;
    ee_unset_lock(&(tnode->parent->lock));
  #endif

  if (thr->num_siblings > 1)   /* lightweight barrier: */
    ort_taskwait(1);           /* basicaly help with any tasks you can find */

#ifdef ORT_DEBUG
  {
    void ort_task_stats(void);
    ort_task_stats();
  }
#endif
  
  __CURRTASK(thr) = tnode->parent;
  ort_task_free(thr, tnode);              /* recycle task node */
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                   *
 * DEBUGGING                                                         *
 *                                                                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifdef ORT_DEBUG

void ort_task_stats(void)
{
  ort_eecb_t *me = __MYCB;
  
  fprintf(stderr, "task stats @ thread %d:\n\t"
                  "             enqueued: %ld\n\t"
                  "  dequeued & executed: %ld\n\t"
                  "                stole: %ld\n\t"
                  "throttled (immediate): %ld\n\t"
                  " {\n\t"
                  "     due to full pool:   %ld\n\t"
                  "    due to full queue:   %ld\n\t"
                  "     due to if(FALSE):   %ld\n\t"
                  "  outside of parallel:   %ld\n\t"
                  "     rest (fast code):   %ld\n\t"
                  " }\n\n",
                  me->thread_num,
                  me->tasking.tasks_enqueued,
                  me->tasking.tasks_executed_by_worker,
                  me->tasking.tasks_executed_by_thief,
	          me->tasking.throttled,
                  me->tasking.throttled_pool,
                  me->tasking.throttled_queue,
                  me->tasking.throttled_if,
	          me->tasking.throttled_serial,
	          me->tasking.throttled - (me->tasking.throttled_pool +
                       me->tasking.throttled_queue +  me->tasking.throttled_if +
	               me->tasking.throttled_serial)
	  );

  me->tasking.tasks_enqueued = 0;
  me->tasking.tasks_executed_by_thief = 0;
  me->tasking.tasks_executed_by_worker = 0;
  me->tasking.throttled = 0;
  me->tasking.throttled_pool = 0;
  me->tasking.throttled_queue = 0;
  me->tasking.throttled_if = 0;
  me->tasking.throttled_serial = 0;
}

#endif
